/*
    File:   asc.c
    Author: Jonathan Dale Kirwan

    Creation Date:  Fri 10-Sep-2004 17:31:40
    Last Modified:  Mon 08-Aug-2005 23:51:06

    Copyright 2004, 2005    Jonathan Dale Kirwan, All Rights Reserved.


    COPYRIGHT NOTICE

    You are granted a non-transferable, non-exclusive, royalty-free worldwide
    license to use, copy, modify, prepare derivative works of and distribute
    this software, subject to your agreement that you acquire no ownership
    right, title, or interest in this software and your agreement that this
    software is research work that is provided 'as is.'

    Jonathan Dale Kirwan disclaims all warranties with regard to this source
    code software, including any and all implied warranties of merchantability
    and fitness of purpose.  In no event shall Jonathan Dale Kirwan be liable
    for any direct, indirect, consequential or special damages or any damages
    whatsoever resulting from loss of use, loss of data or profits, whether in
    an action of contract, negligence or other tortious action, arising out of
    or in connection with the use or performance of this software.

    In other words, you can have my sympathies but that's all you get from me!


    DESCRIPTION

    This program reads schematics that have been captured using Linear Tech's
    LTSpice program (aka Switcher CAD III) and converts them into simple ASCII
    schematics.  The LTSpice program saves its schematics as .ASC files and
    they use legible ASCII tokens to describe the circuit which can be read,
    parsed, and handled by the LTSpice program.  What this program does is to
    read this information and attempt to draw it using ASCII characters, which
    are then simply displayed as text to 'stdout' (for now.)

    I've kept this program to a single module to keep things simpler, as I
    I think having one file is better than several, small functional modules.


    SYNTAX

    The .ASC file has a few different types of lines to process.  The first
    line is always a "Version" line and is the only one with a keyword having
    both upper and lower case.  It's ignored here.  Processed lines will use
    one of the following keywords:

        WIRE FLAG IOPIN SYMBOL SYMATTR WINDOW TEXT RECTANGLE

    Lines that don't start with one of the above keywords aren't processed.
    Some of these include:

        ARC CIRCLE SHEET Version

    There are probably others, of course.  But I've not done a careful study.
    Suffice it that if they aren't in the first list, they aren't processed.

    Some of the processed line types only follow certain other ones and the
    program may (probably does) depend on this observation.  For example, an
    IOPIN line will follow its associated FLAG line and the x and y values
    should match.  Also, SYMATTR and WINDOW lines follow a SYMBOL line, if
    they are present at all.  While it's not entirely necessary that IOPIN
    lines *must* follow their associated FLAG line (since the x and y points
    must match up), the program depends on the observation that such IOPIN
    lines do, in fact, always immediately follow their FLAG line.  The WINDOW
    lines are currently ignored, but the SYMATTR lines must follow their
    SYMBOL lines, as there isn't enough semantic information to otherwise make
    the right associations.


    DESIGN

    There are several ideas I used to figure out an approach.

    I use a GCD algorithm to help me convert abstract schematic wiring points
    in (x,y) form into ASCII (row, column) positions.  By reading the various
    (x, y) values in the .ASC file, I can find the largest common denominator
    among them and then divide them by this value.  That is how I map from an
    (x, y) in LTSpice to a (row, column) for the ASCII schematic.

    LTSpice uses negative wiring points, so there is no assumption made about
    where (0, 0) is, relative to the schematic itself.  However, LTSpice does
    use positive-down, positive-right relative directions in its abstraction,
    so (0,0) will be to the left and above (10,10), for example.  What I then
    do while computing the GCD values to use is to also compute a minimum and
    maximum for x and y so that I can place the corner on the ASCII sheet and
    not clip or excessively grow the resulting ASCII schematic.
    
    I use a special type of character cell that looks something like this:

          ......T......
          .     |     .
          .     |     .
         L.---- + ----.R
          .     |     .
          .     |     .
          ......B......

    The above cell, bordered by periods, must be turned into some ASCII value,
    when the time comes to produce the schematic in ASCII.  But beforehand, it
    is handled as shown above.  Each cell has four, short lines -- a left line
    (L), a right line (R), a top line (T), a bottom line (B), and a connection
    point (+), which is placed at "dotted" points in the schematic.

    I make an X/Y grid of these cells and initialize them as undrawn.  The
    schematic is then drawn onto this grid.  The starting point always places
    a connection point or dot and, if the destination isn't the same cell,
    draws only one of the four lines.  Similarly, the end point.  In those
    cells in between, two lines are drawn and no dot is placed.

    For example, if I observe a WIRE command directing me to draw from (0,1)
    to (2,1) on the ASCII schematic and the current cells in that region look
    like this, beforehand:

                 0           1           2
          .....................................
          .           .           .           .
          .           .           .           .
        0 .           .           .           .
          .           .           .           .
          .           .           .           .
          .....................................
          .           .           .           .
          .           .           .           .
        1 .           .           .           .
          .           .           .           .
          .           .           .           .
          .....................................
          .           .           .           .
          .           .           .           .
        2 .-----------.----       .           .
          .           .           .           .
          .           .           .           .
          .....................................

    Then the resulting region will look, as follows, after the drawing command:

                 0           1           2
          .....................................
          .           .           .           .
          .           .           .           .
        0 .           .     +     .           .
          .           .     |     .           .
          .           .     |     .           .
          .....................................
          .           .     |     .           .
          .           .     |     .           .
        1 .           .           .           .
          .           .     |     .           .
          .           .     |     .           .
          .....................................
          .           .     |     .           .
          .           .     |     .           .
        2 .-----------.---- +     .           .
          .           .           .           .
          .           .           .           .
          .....................................

    Note that dots are placed at the start and end cells and appropriate lines
    are added.

    There are 32 possible permutations, given that a dot can be placed or not
    placed and that each of the four lines can be independently placed or not
    placed, as well.  Not all of these are likely or legal (only some 19 are
    legal, I think) so not all of them need to produce ASCII characters that
    make sense.

    Also, I've set aside another flag (bit) to tell me if the lines and dots
    for a cell should be ignored and, instead, just treat the whole thing as a
    character.  This allows me to place text into the ASCII schematic, thus
    overriding any interpretation for lines and dots.

    That's about it for the details, so far.


    ASCII LIBRARY FILE FORMAT

    This program uses a single ASCII library file to hold the ASCII images for
    each part.  No special characters are used or required, so this library of
    parts can be edited with most any text editor.

    An example looks like:

        ?res 2 -6 0
        ?sym InstName X
        ?sym Value Y
         X
        -/\/\-
         Y
        ?cap 1 0 0
        ?sym InstName X
        ?sym Value Y
         |
        --- X
        --- Y
         |
         |

    The special character, ?, is used to indicate a line that will be parsed
    with an eye to either finding the part name or else 'sym'.  Each part in
    the library file starts with a ?<partname> and has three additional values
    added afterwards.  The fuller format of the line is:

        ?<partname> <orientation> <x-home> <y-home>

    When searching for a part in the library, the program uses the part name
    specified in the LTSpice schematic file and also the rotation/mirror
    orientation given and then searches for the first line that matches this.

    The values are:

        library orientation     LTSpice orientation
                1                     R0
                2                     R90
                3                     R180
                4                     R270
                5                     M0
                6                     M90
                7                     M180
                8                     M270

    When adding new entries to the file, keep the above list in mind.

    The values of <x-home> and <y-home> provide for some minor adjustments of
    the ASCII image that will be placed into the schematic and the origin for
    that part given by LTSpice.  Usually, these two numbers are relatively
    small, but they are often non-zero.  To get them right, just run the
    program with default values (0 or 1, for example, for each) and then see
    how the image gets placed.  If it needs to move left, subtract some from
    <x-home>.  If it needs to move right, add some.  If it needs to move up,
    subtract some from <y-home>.  And, of course, if it needs to move down,
    add some.

    Once the part has been found there, the program then scans for ?sym lines.
    These lines specify LTSpice SYMATTR attributes that will be placed into
    the ASCII image and provides a special character to use in locating them
    correctly.  The format of the ?sym line is:

        ?sym <attribute name> <ASCII replacement character>

    The program goes back to the LTSpice schematic file and reads up the
    associated SYMATTR lines for the part and matches these up with those
    provided by ?sym lines in the library, connecting them on the basis of the
    value in <attribute name>.  Case is important here.  Then, the program
    finds any occurrence of <ASCII replacement character> in the ASCII image
    that follows these ?sym lines with the attributes found in the LTSpice
    schematic file and places the result into the schematic.

    The replacement text is placed left-justified if the <ASCII replacement
    character> is upper case and right-justified, otherwise.

    The actual ASCII schematic follows after the ?sym lines and ends with the
    end of the file or else another line beginning with ?.

    I've also added support for parts to refer to other parts, for display
    purposes.  This means that the first line in the library for a part may
    look like:

        ?npn2,npn

    When the name is followed by a comma and then by a new part name, that new
    name then causes a re-search of the library in order to find the ASCII
    layout.  There's a special ?sym attribute programmed in, called SYMNAME.
    This one substitutes the original part name in cases where a re-search
    took place.  For example, if the symbol for the LT1001 points to 'opamp',
    then LT1001 will replace the character designated for a ?sym SYMNAME line.
    Keep in mind that case is significant, so it must be all upper case here.


    TARGET COMPILER

    This module is designed to be compiled with Microsoft's VC 1.52C compiler
    as either a .COM or a .EXE file and it does depend on some specialized
    library routines as well as the DOS command line.  Please feel free to
    adapt it to other systems (and, if you let me know what you've done, I'd
    be happy to consider adding the changes to the overall source code in
    order to make it more generally compilable.)
 
 
    MODIFICATIONS
 
    Sun 07-Aug-2005 00:42:37    Jon Kirwan                      Version 1.1

        Added several command line options, including +h and -h for
        help, +r and -r for rectangle drawing support (might be
        desired once in a while), and +c and -c for Windows clipboard
        support via WinOldAp.

    Sun 07-Aug-2005 13:53:45    Jon Kirwan                      Version 1.2

        Added the +c<char> capability.  This will automatically prefix
        the given character to each line placed in the clipboard for
        purposes of pasting into newsgroup messages.

    Mon 08-Aug-2005 23:49:51    Jon Kirwan                    Version 1.2.1

        Adjusted the program so that it no longer prints out "" when the
        Value field of various SYMATTR's are empty and thus set to "".
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define MAXBUF      (200)       /* buffer size for reading .ASC files */
#define MAXLST      (20)        /* limit to number of InstName types */

int main( int argc, char* argv[] );


/*  User's options
    --------------------------------------------------------------------------
    This structure is simply a collection place for option settings by the
    application user, which may control the behavior of this program.
*/

typedef struct {            /* --- user's run-time options --- */
    int rect_flag;              /* draw rectangles?  0=no, 1=yes */
    int clip_flag;              /* draw to clipboard?  0=no, 1=yes */
    char clip_char;             /* special char, or \0 if none */
} user_options_t;


/*  ASCII cells
    --------------------------------------------------------------------------
    The various details I need to keep track of, for each ASCII cell, have
    already been described earlier.  However, here are the C declarations I
    use.  Each cell is either a lines_t type or else it is a char type.  What
    decides this is the bit flag, spc.  If the bit is 1, it's a valid lines_t
    type.
*/

typedef struct {            /* --- lines and dots type --- */
    unsigned char dot:1;        /* dotted cell (connection point is placed) */
    unsigned char top:1;        /* top line (T), 1 if present */
    unsigned char bot:1;        /* bottom line (T), 1 if present */
    unsigned char lft:1;        /* left line (T), 1 if present */
    unsigned char rgt:1;        /* right line (T), 1 if present */
    unsigned char :2;
    unsigned char spc:1;        /* 1 if lines_t type, 0 if char */
} lines_t;

typedef struct {            /* --- drawing ID --- */
    unsigned char val:5;        /* lines and dots field value */
    unsigned char :2;
    unsigned char spc:1;        /* 1 if lines_t type, 0 if char */
} draw_t;

typedef union {             /* --- merged character and lines type --- */
    char c;                     /* ASCII character, if not line/dot */
    draw_t d;                   /* field value when making ASCII chars */
    lines_t e;                  /* line/dot type, if not ASCII char */
} tableau_t;


/*  Local Function Prototypes
    --------------------------------------------------------------------------
*/

static char * ReadLine( char *ln, int max, FILE *un );
static char * SymAttr( FILE *f, const char *symname );
static char * ParseASYFileName( const char *s, char **z, int mode );
#define TAIL        (1)         /* parse only the filename tail */
#define FULL        (0)         /* parse the full filename */
static unsigned int gcd( unsigned int a, unsigned int b );
#define igcd(a,b)   ((int) gcd( (unsigned int) abs( a ), (unsigned int) abs( b ) ))
static char * Name( const char *s );
static char * BuildFullPath( const char *fdir, const char *fname, const char *fext );
static void WirePointStats( FILE *fi, int *xgcd, int *ygcd, int *xmin, int *ymin, int *xmax, int *ymax );
static void ParseWIRE( const char *buf, int *x1, int *y1, int *x2, int *y2 );
static char * ParseRECTANGLE( const char *buf, int *ulx, int *uly, int *lrx, int *lry, int *style );
static char * ParseFLAG( const char *buf, int *x1, int *y1 );
static char * ParseTEXT( const char *buf, int *x1, int *y1, char **justify, int *unknown );
static char * ParseSYMBOL( const char *buf, int *x1, int *y1, int *orient );
static char * ParseLibHDR( const char *buf, int *orient, int *x, int *y );
static char * ParseLibSYM( const char *buf, char *c );
static void DrawWIRE( tableau_t *t, int xwidth, int x1, int y1, int x2, int y2 );
static void DrawRECTANGLE( tableau_t *t, int xwidth, int ulx, int uly, int lrx, int lry, int style );
static void DrawFLAG( tableau_t *t, int xwidth, int x, int y, const char *txt );
static void DrawTEXT( tableau_t *t, int xwidth, int x, int y, const char *txt, const char *justify );
static void DrawSYMBOL( FILE *f, tableau_t *t, int xwidth, int x, int y, const char *sym, int orient, FILE *lib );
static tableau_t * CreateASCIIPad( int xwidth, int yheight );
static void DrawASCIIPad( FILE *f, tableau_t *t, int xwidth, int xgcd, int ygcd, int xmin, int ymin, FILE *lib );
static void DisplayASCII( FILE *fi, FILE *fo, FILE *sym );
static int clpVerify( void );
static int clpOpen( void );
static int clpClose( void );
static int clpEmpty( void );
static int clpCopyText( char *txt, unsigned int len );


/*  Module-wide static data.
    --------------------------------------------------------------------------
*/

static char ascii[]= "  *'*,||*-*'*,*+*-*'*,*+--*+*+-+";
static user_options_t options;


/*  ReadLine
    --------------------------------------------------------------------------
    This routine reads a line of text from a file into a specified string.  It
    returns the specified string's pointer, if successful.  If not, it returns
    NULL.  Trailing line control characters (carriage return and linefeed) are
    not included in the returned null-terminated string.
*/

static char * ReadLine( char *ln, int max, FILE *un ) {
    auto int i;
    auto char *p, c;

        /* get a text line from the file */
        if ( fgets( ln, max, un ) == NULL )
            return NULL;

        /* strip trailing carriage returns or linefeeds */
        for ( i= strlen( ln ), p= &ln[i]; i > 0; --i )
            if ( (c= *--p) == '\n' )
                *p= '\0';
            else if ( c == '\r' )
                *p= '\0';
            else
                break;

        /* convert white space into simple spaces */
        for ( p= ln; *p != '\0'; ++p )
            if ( isspace( *p ) )
                *p= ' ';

        /* trim off trailing spaces */
        for ( p= & ln[strlen( ln )]; *--p == ' '; )
            ;
        p[1]= '\0';

    return ln;
}


/*  SymAttr
    --------------------------------------------------------------------------
    This routine examines SYMATTR lines for the indicated symbol type (these
    lines specify a symbol name and some attribute) and extracts the value for
    the given symbol, if found.  The function allocates enough space for the
    resulting value (which must be returned to the heap) and returns the
    pointer.  A NULL value is returned in the case where no matching SYMATTR
    line was discovered.  In all cases, the file pointer is restored to its
    starting position, at exit.

    SYMATTR lines can be interspersed with WINDOW lines, after a SYMBOL line,
    so these WINDOW lines are allowed, but skipped, when searching.  Anything
    other than a SYMATTR or WINDOW line terminates the search.
*/

static char * SymAttr( FILE *f, const char *symname ) {
    auto long pos;
    auto char *aname, *p, *s, buf[MAXBUF];

        s= NULL;
        pos= ftell( f );
        while ( ReadLine( buf, MAXBUF, f ) != NULL )

            /* SYMATTR is parsed for the indicated symbol */
            if ( strncmp( buf, "SYMATTR ", 8 ) == 0 ) {
                aname= ParseASYFileName( buf+8, &p, TAIL );
                if ( strcmp( aname, symname ) == 0 ) {
                    s= (char *) malloc( strlen( p+1 ) + 1 );
                    if ( s != NULL )
                        strcpy( s, p+1 );
                    free( aname );
                    break;
                }
                free( aname );

            /* WINDOW is allowed, but skipped */
            } else if ( strncmp( buf, "WINDOW ", 7 ) == 0 ) {
                
            /* not SYMATTR or WINDOW, so quit */
            } else
                break;

        fseek( f, pos, SEEK_SET );

    return s;
}


/*  ParseASYFileName
    --------------------------------------------------------------------------
    This routine examines the given string and parses a .ASY filename from it,
    allocating an appropriately sized string from the heap and returning the
    string pointer.  The pointer is always non-NULL, unless malloc() couldn't
    allocate the space.  An optional pointer is modified to point to the first
    character that was NOT parsed, if provided as non-NULL.

    This routine handles the backslash escaping mechanism, when parsing the
    .ASY filename.

    The mode is set to 0 (FULL) if the entire filename is to be returned and
    is instead set to 1 (TAIL) if only the local name at the end of it is
    desired (after the last filename slash.)
*/

static char * ParseASYFileName( const char *s, char **z, int mode ) {
    auto char c, *r;
    auto const char *p;

        /* scan the string, adjusting p until non-filename char is seen */
        for (p= s-1; (c= *++p) != '\0'; ) {
            if ( c == '\\' ) {
                if ( (c= *++p) == '\0' )
                    break;
            } else if ( c == ' ' )
                break;
            if ( mode == 1 && c == '\\' )
                s= p+1;
        }

        /* return the pointer to the first unprocessed character */
        if ( z != NULL )
            *z= (char *) p;

        /* allocate space for the filename/tail and return it */
        if ( (r= (char *) malloc ( (size_t) (p - s + 1) )) != NULL ) {
            strncpy( r, s, p - s );
            r[p-s]= '\0';
        }

    return r;
}


/*  gcd
    --------------------------------------------------------------------------
    This routine returns the greatest common divisor, given two integers.  If
    either of them are zero, then the value of the other is returned.  If both
    are zero, zero is returned.
    
    To keep it simple, I've not bothered using divison here.
*/

static unsigned int gcd( unsigned int a, unsigned int b )
{
        if ( a == 0 && b == 0 )
            ;
        else if ( b == 0 )
            b= a;
        else if ( a != 0 )
            while ( a != b )
                if ( a < b )
                    b -= a;
                else
                    a -= b;

    return b;
}


/*  Name
    --------------------------------------------------------------------------
    This routine accepts argv[0] (usually) and extracts the program name part,
    for use with error messages.  The string space for the returned string is
    allocated from the heap and must be returned, later.
*/

static char * Name( const char *s ) {
    auto char dv[_MAX_DRIVE], dr[_MAX_DIR], fn[_MAX_FNAME], ex[_MAX_EXT];
    auto size_t slen;
    auto char *r;

        /* split up the provided full path into file name parts */
        _splitpath( s, dv, dr, fn, ex );

        /* substitute a default name, if none given in the path */
        slen= strlen( fn );
        if ( slen < 1 ) {
            strcpy( fn, "ASCSCHEM" );
            slen= strlen( fn );
        }

        /* allocate and copy the string */
        r= (char*) malloc( slen + 1 );
        strcpy( r, fn );

        /* convert to lower case */
        _strlwr( r );

    return r;
}


/*  BuildFullPath
    --------------------------------------------------------------------------
    This routine accepts a directory (which may also include a file name and
    extension that will be ignored) and a file name and extension and builds a
    new full path name using them in combination.  The string space for the
    returned string is allocated from the heap and must be returned, later.
*/

static char * BuildFullPath( const char *fdir, const char *fname, const char *fext ) {
    auto char pth[_MAX_PATH], dv[_MAX_DRIVE], dr[_MAX_DIR], fn[_MAX_FNAME], ex[_MAX_EXT];
    auto char *r;

        /* split up the provided full path into file name parts */
        _splitpath( fdir, dv, dr, fn, ex );

        /* use parts from the original path, if specified as NULL */
        if ( fname == NULL )
            fname= fn;
        if ( fext == NULL )
            fext= ex;

        /* rebuild the full path name */
        _makepath( pth, dv, dr, fname, fext );

        /* allocate and copy the string */
        r= (char *) malloc( strlen( pth ) + 1 );
        strcpy( r, pth );

    return r;
}


/*  ParseWIRE
    --------------------------------------------------------------------------
    Parses the WIRE line in a .ASC schematic file.  There are four values that
    represent the starting and ending (x, y) points.
*/

static void ParseWIRE( const char *buf, int *x1, int *y1, int *x2, int *y2 ) {
    auto char *p;

        *x1= (int) strtol( buf+5, &p, 0 );
        *y1= (int) strtol( p+1, &p, 0 );
        *x2= (int) strtol( p+1, &p, 0 );
        *y2= (int) strtol( p+1, &p, 0 );

    return;
}


/*  ParseRECTANGLE
    --------------------------------------------------------------------------
    Parses the RECTANGLE drawing shape in a .ASC schematic file.  There is a
    value which is usually 'Normal', four values that represent the upper left
    and lower right corners as separate (x, y) points, and then a line style
    which is a numeric value or else is missing entirely.  If missing, the
    line style is assigned the value of 0.
*/

static char * ParseRECTANGLE( const char *buf, int *ulx, int *uly, int *lrx, int *lry, int *style ) {
    auto char *p, *n, *z;

        n= ParseASYFileName( buf+10, &p, TAIL );   /* used for convenience */
        *ulx= (int) strtol( p+1, &p, 0 );
        *uly= (int) strtol( p+1, &p, 0 );
        *lrx= (int) strtol( p+1, &p, 0 );
        *lry= (int) strtol( p+1, &p, 0 );
        *style= (int) strtol( p+1, &z, 0 );
        if ( z == p + 1 )
            *style= 0;

    return n;
}


/*  ParseFLAG
    --------------------------------------------------------------------------
    Parses the FLAG line in a .ASC schematic file.  There are two values that
    represent the flag's (x, y) position.  Also, the return value holds the
    flag's text in a string that is allocated from the heap (and must be
    returned to the heap.)
*/

static char * ParseFLAG( const char *buf, int *x, int *y ) {
    auto char *p, *n;

        *x= (int) strtol( buf+5, &p, 0 );
        *y= (int) strtol( p+1, &p, 0 );
        n= (char *) malloc( strlen( p+1 ) + 1 );
        if ( n != NULL )
            strcpy( n, p+1 );

    return n;
}


/*  ParseTEXT
    --------------------------------------------------------------------------
    Parses the TEXT line in a .ASC schematic file.  There are two values that
    represent the text's (x, y) position.  Also, the return value holds the
    display text in a string that is allocated from the heap (and must be
    returned to the heap.)  Finally, justify is also modified to provide the
    type of justification (Left, Right, etc.)
*/

static char * ParseTEXT( const char *buf, int *x, int *y, char **justify, int *unknown ) {
    auto char *p, *n;

        *x= (int) strtol( buf+5, &p, 0 );
        *y= (int) strtol( p+1, &p, 0 );
        *justify= ParseASYFileName( p+1, &p, TAIL );
        *unknown= (int) strtol( p+1, &p, 0 );
        while ( *p == ' ' )
            ++p;
        if ( *p == '!' || *p == ';' )
            ++p;
        n= (char *) malloc( strlen( p ) + 1 );
        if ( n != NULL )
            strcpy( n, p );

    return n;
}


/*  ParseSYMBOL
    --------------------------------------------------------------------------
    Parses the SYMBOL line in a .ASC schematic file.  There are two values
    that represent the symbol's (x, y) corner position.  Also, the return
    value holds the symbol's name in a string that is allocated from the heap
    (and must be returned to the heap.)
*/

static char * ParseSYMBOL( const char *buf, int *x, int *y, int *orient ) {
    auto char *p, *n;

        n= ParseASYFileName( buf+7, &p, TAIL );
        *x= (int) strtol( p+1, &p, 0 );
        *y= (int) strtol( p+1, &p, 0 );
        *orient= 1 + ((int) strtol( p+2, NULL, 0 )) / 90;
        if ( p[1] == 'M' )
            *orient += 4;

    return n;
}


/*  ParseLibHDR
    --------------------------------------------------------------------------
    Parses a symbol library header line.  These are formatted with the symbol
    name followed by the symbol orientation and home position's x and y, all
    separated by spaces.
*/

static char * ParseLibHDR( const char *buf, int *orient, int *x, int *y ) {
    auto char *p, *n;

        n= ParseASYFileName( buf, &p, TAIL );   /* used for convenience */
        *orient= (int) strtol( p+1, &p, 0 );
        *x= (int) strtol( p+1, &p, 0 );
        *y= (int) strtol( p+1, &p, 0 );

    return n;
}


/*  ParseLibSYM
    --------------------------------------------------------------------------
    Parses a symbol library symbol replacement line.  These are formatted with
    replacement symbol name followed by the single character ID that will be
    used in the symbol's display text and will be replaced, when seen.
*/

static char * ParseLibSYM( const char *buf, char *c ) {
    auto char *p, *n;

        free( ParseASYFileName( buf, &p, TAIL ) );  /* throw away keyword */
        n= ParseASYFileName( p+1, &p, TAIL );       /* get LTSpice symbol */
        *c= p[1];

    return n;
}


/*  WirePointStats
    --------------------------------------------------------------------------
    This routine rewinds and then reads the indicated file and computes the
    minimum and maximum x and y values, as well as the gcd for x and y.  Only
    WIRE, FLAG, and SYMBOL line types are processed here.
*/

static void WirePointStats( FILE *fi, int *xgcd, int *ygcd, int *xmin, int *ymin, int *xmax, int *ymax ) {
    auto int flg;
    auto char buf[MAXBUF];

        flg= 0;
        *xgcd= *ygcd= 0;

        fseek( fi, 0L, SEEK_SET );

        while ( ReadLine( buf, MAXBUF, fi) != NULL ) {

            if ( strncmp( buf, "WIRE ", 5 ) == 0 ) {
              auto int x1, y1, x2, y2;
                ParseWIRE( buf, &x1, &y1, &x2, &y2 );
                *xgcd= igcd( igcd( *xgcd, x1 ), x2 );
                *ygcd= igcd( igcd( *ygcd, y1 ), y2 );
                if ( flg == 0 ) {
                    *xmin= *xmax= x1;
                    *ymin= *ymax= y1;
                    flg= 1;
                }
                *xmin= __min( __min( *xmin, x1 ), x2 );
                *xmax= __max( __max( *xmax, x1 ), x2 );
                *ymin= __min( __min( *ymin, y1 ), y2 );
                *ymax= __max( __max( *ymax, y1 ), y2 );
            }

            else if ( strncmp( buf, "FLAG ", 5 ) == 0 ) {
              auto int x, y;
              auto char *txt;
                txt= ParseFLAG( buf, &x, &y );
                *xgcd= igcd( *xgcd, x );
                *ygcd= igcd( *ygcd, y );
                if ( flg == 0 ) {
                    *xmin= *xmax= x;
                    *ymin= *ymax= y;
                    flg= 1;
                }
                else {
                    *xmin= __min( *xmin, x );
                    *xmax= __max( *xmax, x );
                    *ymin= __min( *ymin, y );
                    *ymax= __max( *ymax, y );
                }
                free( txt );
            }

            else if ( strncmp( buf, "TEXT ", 5 ) == 0 ) {
              auto int x, y, unknown;
              auto char *justify, *txt;
                txt= ParseTEXT( buf, &x, &y, &justify, &unknown );
                if ( flg == 1 ) {
                  auto int m;
                    if ( x < 0 ) {
                        m= (-x) % *xgcd;
                        if ( m != 0 )
                            x += (m - *xgcd);
                    }
                    else {
                        m= x % *xgcd;
                        if ( m != 0 )
                            x -= (m - *xgcd);
                    }
                    if ( y < 0 ) {
                        m= (-y) % *ygcd;
                        if ( m != 0 )
                            y += (m - *ygcd);
                    }
                    else {
                        m= y % *ygcd;
                        if ( m != 0 )
                            y -= (m - *ygcd);
                    }
                    *xmin= __min( *xmin, x );
                    *xmax= __max( *xmax, x );
                    *ymin= __min( *ymin, y );
                    *ymax= __max( *ymax, y );
                }
                free( justify );
                free( txt );
            }

            else if ( strncmp( buf, "SYMBOL ", 7 ) == 0 ) {
              auto int x, y;
              auto char *sym;
              auto int orient;
                sym= ParseSYMBOL( buf, &x, &y, &orient );
                *xgcd= igcd( *xgcd, x );
                *ygcd= igcd( *ygcd, y );
                if ( flg == 0 ) {
                    *xmin= *xmax= x;
                    *ymin= *ymax= y;
                    flg= 1;
                }
                else {
                    *xmin= __min( *xmin, x );
                    *xmax= __max( *xmax, x );
                    *ymin= __min( *ymin, y );
                    *ymax= __max( *ymax, y );
                }
                free( sym );
            }

            else if ( options.rect_flag && strncmp( buf, "RECTANGLE ", 10 ) == 0 ) {
              auto int ulx, uly, lrx, lry;
              auto char *sym;
              auto int style;
                sym= ParseRECTANGLE( buf, &ulx, &uly, &lrx, &lry, &style );
                *xgcd= igcd( igcd( *xgcd, ulx ), lrx );
                *ygcd= igcd( igcd( *ygcd, uly ), lry );
                if ( flg == 0 ) {
                    *xmin= *xmax= ulx;
                    *ymin= *ymax= uly;
                    flg= 1;
                }
                *xmin= __min( __min( *xmin, ulx ), lrx );
                *xmax= __max( __max( *xmax, ulx ), lrx );
                *ymin= __min( __min( *ymin, uly ), lry );
                *ymax= __max( __max( *ymax, uly ), lry );
                free( sym );
            }
        }

    return;
}


/*  DrawWIRE
    --------------------------------------------------------------------------
    This routine draws a wire line on the ASCII schematic pad.
*/

static void DrawWIRE( tableau_t *t, int xwidth, int x1, int y1, int x2, int y2 ) {

        /* handle case where the start point and end point are the same */
        if ( x1 == x2 && y1 == y2 )
            t[x1+xwidth*y1].e.dot= 1;

        /* handle case where the line is vertical */
        else if ( x1 == x2 ) {
          auto int y;
            t[x1+xwidth*y1].e.dot= 1;
            t[x1+xwidth*y2].e.dot= 1;
            if ( y1 > y2 ) {
              auto int tmp;
                tmp= y1;
                y1= y2;
                y2= tmp;
            }
            t[x1+xwidth*y1].e.bot= 1;
            t[x1+xwidth*y2].e.top= 1;
            for ( y= y1+1; y < y2; ++y ) {
                t[x1+xwidth*y].e.top= 1;
                t[x1+xwidth*y].e.bot= 1;
            }

        /* handle case where the line is horizontal */
        } else if ( y1 == y2 ) {
          auto int x;
            t[x1+xwidth*y1].e.dot= 1;
            t[x2+xwidth*y1].e.dot= 1;
            if ( x1 > x2 ) {
              auto int tmp;
                tmp= x1;
                x1= x2;
                x2= tmp;
            }
            t[x1+xwidth*y1].e.rgt= 1;
            t[x2+xwidth*y1].e.lft= 1;
            for ( x= x1+1; x < x2; ++x ) {
                t[x+xwidth*y1].e.lft= 1;
                t[x+xwidth*y1].e.rgt= 1;
            }
        }

        /* other cases are ignored -- slanted lines */

    return;
}


/*  DrawRECTANGLE
    --------------------------------------------------------------------------
    This routine draws a rectangle on the ASCII schematic pad.
*/

static void DrawRECTANGLE( tableau_t *t, int xwidth, int ulx, int uly, int lrx, int lry, int style ) {

        DrawWIRE( t, xwidth, ulx, uly, ulx, lry );
        DrawWIRE( t, xwidth, ulx, lry, lrx, lry );
        DrawWIRE( t, xwidth, lrx, lry, lrx, uly );
        DrawWIRE( t, xwidth, lrx, uly, ulx, uly );

    return;
}


/*  DrawFLAG
    --------------------------------------------------------------------------
    This routine draws a flag (named node) on the ASCII schematic pad.  For
    now, this routine assumes that there is enough room on the tableau for the
    flag (it doesn't check to see if the text would proceed off some edge.)
*/

static void DrawFLAG( tableau_t *t, int xwidth, int x, int y, const char *txt ) {
    auto const char *p;
    auto char c;
    auto int idx= x+xwidth*y;

        if ( strcmp( txt, "0" ) == 0 )
            p= "gnd";
        else
            p= txt;

        /* vertical wire passing through places the flag parallel */
        if ( t[idx].e.top == 1 && t[idx].e.bot == 1 ) {
            --x;
            for ( y -= (strlen( p ) / 2); (c= *p) != '\0'; ++y, ++p )
                t[x+xwidth*y].c= c;

        /* horizontal wire passing through places the flag parallel */
        } else if ( t[idx].e.lft == 1 && t[idx].e.rgt == 1 ) {
            --y;
            for ( x -= (strlen( p ) / 2); (c= *p) != '\0'; ++x, ++p )
                t[x+xwidth*y].c= c;

        /* vertical wire stubbing down places the flag below */
        } else if ( t[idx].e.top == 1 ) {
            t[idx].e.bot= 1;
            ++y;
            for ( x -= (strlen( p ) / 2); (c= *p) != '\0'; ++x, ++p )
                t[x+xwidth*y].c= c;

        /* vertical wire stubbing up places the flag above */
        } else if ( t[idx].e.bot == 1 ) {
            t[idx].e.top= 1;
            --y;
            for ( x -= (strlen( p ) / 2); (c= *p) != '\0'; ++x, ++p )
                t[x+xwidth*y].c= c;

        /* horizontal wire from the left places the flag to right */
        } else if ( t[idx].e.lft == 1 ) {
            t[x+xwidth*y].e.rgt= 1;
            for ( ++x; (c= *p) != '\0'; ++x, ++p )
                t[x+xwidth*y].c= c;

        /* horizontal wire from the right places the flag to left */
        } else if ( t[idx].e.rgt == 1 ) {
            t[x+xwidth*y].e.lft= 1;
            for ( x -= strlen( p ); (c= *p) != '\0'; ++x, ++p )
                t[x+xwidth*y].c= c;
        }

        /* and regardless, dot the cell */
        t[idx].e.dot= 1;

    return;
}


/*  DrawTEXT
    --------------------------------------------------------------------------
    This routine draws a flag (named node) on the ASCII schematic pad.  For
    now, this routine assumes that there is enough room on the tableau for the
    flag (it doesn't check to see if the text would proceed off some edge.)
*/

static void DrawTEXT( tableau_t *t, int xwidth, int x, int y, const char *txt, const char *justify ) {
    auto char c;
    auto int xmargin;

        for ( xmargin= x; (c= *txt) != '\0'; ++x, ++txt ) {
            if ( c == '\\' ) {
                c= *++txt;
                if ( c == '\0' ) {
                    t[x+xwidth*y].c= '\\';
                    break;
                }
                else if ( c == 'n' ) {
                    ++y;
                    x= xmargin-1;
                }
                else {
                    t[x+xwidth*y].c= '\\';
                    ++x;
                    t[x+xwidth*y].c= c;
                }
            }
            else
                t[x+xwidth*y].c= c;
        }

    return;
}


/*  DrawSYMBOL
    --------------------------------------------------------------------------
    This routine draws a named symbol on the ASCII schematic pad.  These
    symbols are found in the indicated 'library' file.
*/

static void DrawSYMBOL( FILE *f, tableau_t *t, int xwidth, int x, int y, const char *sym, int orient, FILE *lib ) {
    auto char buf[MAXBUF], lstc[MAXLST], *lstnam[MAXLST];
    auto const char *srch;
    auto long libseek;
    auto int i, j, xhome, yhome, liborient, lstcnt, srchlen;

        /*  find the symbol in the library -- in order to allow a library item
            to point to another one for display purposes, this search process
            is iterative.  if an entry refers to another, the library is then
            re-searched using this new name.  the process repeats until the
            entry does NOT refer to another part.
        */
        srch= sym;
restart:
        srchlen= strlen( srch );
        fseek( lib, 0L, SEEK_SET );
        for ( libseek= 0L; ReadLine( buf, MAXBUF, lib ) != NULL; ) {
            if ( buf[0] == '?' && _strnicmp( buf+1, srch, srchlen ) == 0 ) {
                if ( buf[1+srchlen] == ',' ) {
                  auto char *tmp;
                    tmp= (char *) malloc( strlen( buf ) - srchlen - 1 );
                    if ( tmp == NULL )
                        break;
                    strcpy( tmp, buf+srchlen+2 );
                    if ( srch != sym )
                        free( (void *) srch );
                    srch= tmp;
                    goto restart;
                } else if ( buf[1+srchlen] == ' ' ) {
                    free( ParseLibHDR( buf, &liborient, &xhome, &yhome ) );
                    if ( liborient == orient ) {
                        libseek= ftell( lib );
                        break;
                    }
                }
            }
        }
        if ( libseek == 0L )
            return;

        /* build a list of replacement symbols -- which may be empty */
        for ( lstcnt= 0; ReadLine( buf, MAXBUF, lib ) != NULL; ) {
            if ( _strnicmp( buf, "?sym", 4 ) == 0 ) {
                lstnam[lstcnt]= ParseLibSYM( buf, &lstc[lstcnt] );
                ++lstcnt;
                libseek= ftell( lib );
            } else {
                fseek( lib, libseek, SEEK_SET );
                break;
            }
        }

        /* draw and replace */
        for ( j= 0; ReadLine( buf, MAXBUF, lib ) != NULL; ++j ) {
          auto char c;
            if ( buf[0] == '?' )
                break;
            for ( i= 0; (c= buf[i]) != '\0'; ++i )
                if ( c != ' ' ) {
                  auto int s;
                  auto int idx= (x+xhome+i) + xwidth * (y+yhome+j);
                    for ( s= 0; s < lstcnt; ++s )
                        if ( c == lstc[s] )
                            break;
                    if ( s < lstcnt ) {
                      auto char *a;
                        a= SymAttr( f, lstnam[s] );
                        if ( a == NULL && sym != srch && strcmp( lstnam[s], "SYMNAME" ) == 0 ) {
                            a= (char *) malloc( strlen( sym ) + 1 );
                            if ( a != NULL )
                                strcpy( a, sym );
                        }
                        /* if it matches, display the replacement text */
                        if ( a != NULL ) {
                          auto int k;
                          auto unsigned char ac;
                            /* handle "" as an empty liternal and, so, do not print it */
                            if ( strcmp( a, "\"\"" ) == 0 )
                                ;
                            /* handle left justified case */
                            else if ( !islower( c ) ) {
                                for ( k= 0; (ac= (unsigned char) a[k]) != '\0'; ++k ) {
                                    if ( ac == 0xB5 )
                                        ac= 'u';
                                    if ( x + xhome + i + k >= xwidth )
                                        break;
                                    if ( t[idx+k].e.spc == 1 ) {
                                        if ( ascii[t[idx+k].d.val] != ' ' )
                                            break;
                                    } else {
                                        if ( t[idx+k].c != ' ' )
                                            break;
                                    }
                                    t[idx+k].c= (char) ac;
                                }
                            /* handle right justified case */
                            } else {
                              auto int alen= strlen( a );
                                for ( k= 0; (ac= (unsigned char) a[k]) != '\0'; ++k ) {
                                    if ( ac == 0xB5 )
                                        ac= 'u';
                                    if ( x + xhome + i - alen + k + 1 < 0 )
                                        continue;
                                    if ( t[idx-alen+k+1].e.spc == 1 ) {
                                        if ( ascii[t[idx-alen+k+1].d.val] != ' ' )
                                            break;
                                    } else {
                                        if ( t[idx-alen+k+1].c != ' ' )
                                            break;
                                    }
                                    t[idx-alen+k+1].c= (char) ac;
                                }
                            }
                            free( a );
                        }
                        /* otherwise just ignore the whole thing */
                    } else if ( c == '-' ) {
                        if ( t[idx].e.spc == 1 ) {
                            t[idx].e.dot= 1;
                            if ( t[idx].e.top == 1 || t[idx].e.bot == 1 ) {
                                if ( i == 0 )
                                    t[idx].e.rgt= 1;
                                else
                                    t[idx].e.lft= 1;
                            } else
                                t[idx].e.lft= t[idx].e.rgt= 1;
                        } else
                            t[idx].c= c;
                    } else if ( c == '|' ) {
                        if ( t[idx].e.spc == 1 ) {
                            t[idx].e.dot= 1;
                            if ( t[idx].e.lft == 1 || t[idx].e.rgt == 1 ) {
                                if ( j == 0 )
                                    t[idx].e.bot= 1;
                                else
                                    t[idx].e.top= 1;
                            } else
                                t[idx].e.top= t[idx].e.bot= 1;
                        } else
                            t[idx].c= c;
                    } else
                        t[idx].c= c;
                }
        }

        /* free up the list of replacement symbols */
        for ( i= 0; i < lstcnt; ++i )
            free( lstnam[i] );

        /* free up redirected name, if so */
        if ( srch != sym )
            free( (void *) srch );

    return;
}


/*  CreateASCIIPad
    --------------------------------------------------------------------------
    This is the main conversion routine that accepts an input file stream and
    reads it as an LTSpice schematic, processes it, and then writes the ASCII
    schematic to the given output file stream.
*/

static tableau_t * CreateASCIIPad( int xwidth, int yheight ) {
    auto size_t i;
    auto tableau_t *wp;

        wp= (tableau_t *) calloc( (size_t) yheight * xwidth, sizeof( tableau_t ) );
        if ( wp != NULL ) {
            for ( i= yheight * xwidth; i > 0; --i )
                wp[i-1].e.spc= 1;
        }

    return wp;
}


/*  DrawASCIIPad
    --------------------------------------------------------------------------
    This routine reads up each line in the LTSpice schematic and executes the
    necessary drawing commands on the ASCII tableau.
*/

static void DrawASCIIPad( FILE *f, tableau_t *t, int xwidth, int xgcd, int ygcd, int xmin, int ymin, FILE *lib ) {
    auto char buf[MAXBUF];

        fseek( f, 0L, SEEK_SET );
        while ( ReadLine( buf, MAXBUF, f ) != NULL ) {

            /* WIRE directive */
            if ( strncmp( buf, "WIRE ", 5 ) == 0 ) {
              auto int x1, y1, x2, y2;
                ParseWIRE( buf, &x1, &y1, &x2, &y2 );
                x1= x1 / xgcd - xmin;
                y1= y1 / ygcd - ymin;
                x2= x2 / xgcd - xmin;
                y2= y2 / ygcd - ymin;
                DrawWIRE( t, xwidth, x1, y1, x2, y2 );
            }

            /* FLAG directive */
            else if ( strncmp( buf, "FLAG ", 5 ) == 0 ) {
              auto char *txt;
              auto int x, y;
                txt= ParseFLAG( buf, &x, &y );
                x= x / xgcd - xmin;
                y= y / ygcd - ymin;
                DrawFLAG( t, xwidth, x, y, txt );
                free( txt );
            }

            /* SYMBOL directive */
            else if ( strncmp( buf, "SYMBOL ", 7 ) == 0 ) {
              auto char *sym;
              auto int x, y, orient;
                sym= ParseSYMBOL( buf, &x, &y, &orient );
                x= x / xgcd - xmin;
                y= y / ygcd - ymin;
                DrawSYMBOL( f, t, xwidth, x, y, sym, orient, lib );
                free( sym );
            }

            /* TEXT directive */
            else if ( strncmp( buf, "TEXT ", 5 ) == 0 ) {
              auto char *justify, *txt;
              auto int x, y, unknown;
                txt= ParseTEXT( buf, &x, &y, &justify, &unknown );
                if ( (x % xgcd) != 0 )
                    x= 1 + x / xgcd - xmin;
                else
                    x= x / xgcd - xmin;
                y= y / ygcd - ymin;
                DrawTEXT( t, xwidth, x, y, txt, justify );
                free( justify );
                free( txt );
            }

            /* RECTANGLE directive */
            else if ( options.rect_flag && strncmp( buf, "RECTANGLE ", 10 ) == 0 ) {
              auto char *sym;
              auto int ulx, uly, lrx, lry, style;
                sym= ParseRECTANGLE( buf, &ulx, &uly, &lrx, &lry, &style );
                ulx= ulx / xgcd - xmin;
                uly= uly / ygcd - ymin;
                lrx= lrx / xgcd - xmin;
                lry= lry / ygcd - ymin;
                DrawRECTANGLE( t, xwidth, ulx, uly, lrx, lry, style );
                free( sym );
            }
        }

    return;
}


/*  DisplayASCII
    --------------------------------------------------------------------------
    This is the main conversion routine that accepts an input file stream and
    reads it as an LTSpice schematic, processes it, and then writes the ASCII
    schematic to the given output file stream.

    Leading blank lines, trailing blank lines, leading spaces on the left, and
    trailing spaces at the end are all removed.
*/

static void DisplayASCII( FILE *fi, FILE *fo, FILE *sym ) {
    auto tableau_t *wp;
    auto int xgcd, ygcd, xmin, ymin, xmax, ymax;
    auto int xwidth, yheight, x, y, ibase, lines, trimleft;
    auto unsigned int count;

        /* First pass -- process (x, y) points for GCD, min, and max */
        WirePointStats( fi, &xgcd, &ygcd, &xmin, &ymin, &xmax, &ymax );
        if ( xgcd == 0 || ygcd == 0 )
            return;

        /* compute the min and max ASCII character coordinates */
        xmin /= xgcd;
        xmax /= xgcd;
        ymin /= ygcd;
        ymax /= ygcd;

        /* put some padded margins on the left, right, top and bottom */
        xmin -= 20;
        ymin -= 20;
        xmax += 20;
        ymax += 20;

        /* compute the size of the ASCII schematic tableau and allocate it */
        xwidth= xmax - xmin + 1;
        yheight= ymax - ymin + 1;
        wp= CreateASCIIPad( xwidth, yheight );
        if (wp == NULL )
            return;

        /* Second pass -- draw the LTSpice schematic onto this tableau */
        DrawASCIIPad( fi, wp, xwidth, xgcd, ygcd, xmin, ymin, sym );

        /* find least number of leading spaces in the ASCII schematic */
        trimleft= xwidth;
        for ( ibase= 0, y= 0; y < yheight; ibase += xwidth, ++y ) {
          auto int blanks= 0;
            for ( x= 0; x < xwidth; ++x ) {
                if ( (wp[ibase+x].e.spc == 0 && wp[ibase+x].c == ' ') ||
                     (wp[ibase+x].e.spc == 1 && ascii[wp[ibase+x].d.val] == ' ') )
                    ++blanks;
                else {
                    if ( blanks < trimleft )
                        trimleft= blanks;
                    break;
                }
            }
        }

#define outc(a,b)   do { fputc( (a), (b) ); ++count; } while (0)
#define outb(b)     do { if ( options.clip_char != '\0' ) ++count; } while (0)
#define oute(b)     do { fputc( '\n', (b) ); count += 2; } while (0)

        /* display the tableau as ASCII output */
        count= 0;
        lines= -1;
        for ( ibase= 0, y= 0; y < yheight; ibase += xwidth, ++y ) {
          auto int blanks= 0, empty= 1;
            for ( x= trimleft-1; x < xwidth; ++x ) {
              auto char c;
                if ( wp[ibase+x].e.spc == 0 ) {
                    if ( (c= wp[ibase+x].c) == ' ' )
                        ++blanks;
                }
                else if ( (c= ascii[wp[ibase+x].d.val]) == ' ' )
                    ++blanks;
                if ( c != ' ' ) {
                    if ( lines != 0 ) {
                        for ( ; lines > 0; --lines ) {
                            outb( fo );
                            oute( fo );
                        }
                        lines= 0;
                    }
                    if ( empty ) {
                        outb( fo );
                        empty= 0;
                    }
                    for ( ; blanks > 0; --blanks )
                        outc( ' ', fo );
                    outc( c, fo );
                }
            }
            if ( empty ) {
                if ( lines >= 0 )
                    ++lines;
            } else      /* lines guaranteed to be 0 here */
                oute( fo );
        }

#undef outc
#undef outb
#undef oute
#define outc(a,b)   do { *p++= (a); } while (0)
#define outb(b)     do { if ( options.clip_char != '\0' ) *p++= options.clip_char; } while (0)
#define oute(b)     do { *p++= '\r'; *p++= '\n'; } while (0)

        /* display the tableau to the clipboard */
        if ( options.clip_flag ) {
          auto char *cliptxt, *p;
            cliptxt= (char *) malloc( count + 1 );
            if ( cliptxt ) {
                p= cliptxt;
                lines= -1;
                for ( ibase= 0, y= 0; y < yheight; ibase += xwidth, ++y ) {
                  auto int blanks= 0, empty= 1;
                    for ( x= trimleft-1; x < xwidth; ++x ) {
                      auto char c;
                        if ( wp[ibase+x].e.spc == 0 ) {
                            if ( (c= wp[ibase+x].c) == ' ' )
                                ++blanks;
                        }
                        else if ( (c= ascii[wp[ibase+x].d.val]) == ' ' )
                            ++blanks;
                        if ( c != ' ' ) {
                            if ( lines != 0 ) {
                                for ( ; lines > 0; --lines ) {
                                    outb( fo );
                                    oute( fo );
                                }
                                lines= 0;
                            }
                            if ( empty ) {
                                outb( fo );
                                empty= 0;
                            }
                            for ( ; blanks > 0; --blanks )
                                outc( ' ', fo );
                            outc( c, fo );
                        }
                    }
                    if ( empty ) {
                        if ( lines >= 0 )
                            ++lines;
                    } else      /* lines guaranteed to be 0 here */
                        oute( fo );
                }
                *p= '\0';
                if ( clpOpen( ) ) {
                    (void) clpEmpty( );
                    (void) clpCopyText( cliptxt, count );
                    (void) clpClose( );
                }
                free( (void *) cliptxt );
            }
        }

        free( (void *) wp );

    return;
}


/*  clpVerify
    --------------------------------------------------------------------------
    This routine validates the presence of WinOldAp.  WinOldAp supports access
    to the clipboard from DOS.  It returns 1 if WinOldAp is present, otherwise
    0.
*/

#pragma warning( disable:4704 )
static int clpVerify( void ) {
    auto unsigned char minor, major;
    auto unsigned int result;

        __asm {
            mov ax, 1700h
            int 2Fh
            mov major, al
            mov minor, ah
            mov result, ax
        }

    return result != 0x1700;
}
#pragma warning( default:4704 )


/*  clpOpen
    --------------------------------------------------------------------------
    This routine opens the clipboard via WinOldAp.  If successful, it returns
    1.  If not, 0.
*/

#pragma warning( disable:4704 )
static int clpOpen( void ) {
    auto unsigned int result;

        __asm {
            mov ax, 1700h
            int 2Fh
            mov result, ax
        }
        if ( result == 0x1700 )
            return 0;

        __asm {
            mov ax, 1701h
            int 2Fh
            mov result, ax
        }

    return result != 0;
}
#pragma warning( default:4704 )


/*  clpClose
    --------------------------------------------------------------------------
    This routine closes the clipboard via WinOldAp.  If successful, it returns
    1.  If not, 0.
*/

#pragma warning( disable:4704 )
static int clpClose( void ) {
    auto unsigned int result;

        __asm {
            mov ax, 1700h
            int 2Fh
            mov result, ax
        }
        if ( result == 0x1700 )
            return 0;

        __asm {
            mov ax, 1708h
            int 2Fh
            mov result, ax
        }

    return result != 0;
}
#pragma warning( default:4704 )


/*  clpEmpty
    --------------------------------------------------------------------------
    This routine empties the clipboard via WinOldAp.  If successful, it
    returns 1.  If not, 0.
*/

#pragma warning( disable:4704 )
static int clpEmpty( void ) {
    auto unsigned int result;

        __asm {
            mov ax, 1700h
            int 2Fh
            mov result, ax
        }
        if ( result == 0x1700 )
            return 0;

        __asm {
            mov ax, 1702h
            int 2Fh
            mov result, ax
        }

    return result != 0;
}
#pragma warning( default:4704 )


/*  clpCopyText
    --------------------------------------------------------------------------
    This routine places the given text onto the Windows clipboard via WinOldAp
    and returns 1, if successful.  If not, it returns 0.
*/

#pragma warning( disable:4704 )
static int clpCopyText( char *txt, unsigned int len ) {
    auto unsigned int result;
    auto char __far * ptxt;

        __asm {
            mov ax, 1700h
            int 2Fh
            mov result, ax
        }
        if ( result == 0x1700 )
            return 0;

        ptxt= (char __far *) txt;
        __asm {
            mov ax, 1703h
            mov dx, 1
            les bx, ptxt
            push si
            xor si, si
            mov cx, len
            int 2Fh
            pop si
            mov result, ax
        }

    return result != 0;
}
#pragma warning( default:4704 )


/*  main
    --------------------------------------------------------------------------
    This is where the program starts!  If there is no command line argument is
    provided, then stdin is assumed.  And in that case or in the case where a
    single file is provided, stdout is assumed for the resulting output.  In
    all other cases, where more than one file is specified, output is written
    to the same filename, except that the extension used is .TXT.
*/

int main( int argc, char* argv[] ) {
    auto char *n, *sn;
    auto FILE *snunit;
    auto int i, help_flag;

        /* get the name of this program */
        n= Name( argv[0] );

        /* specify the name of the ASCII symbol file and open it */
        sn= BuildFullPath( argv[0], NULL, "SYM" );
        snunit= fopen( sn, "r" );
        ++argv;
        --argc;

        /* Handle switch options */
        options.rect_flag= 0;
        options.clip_flag= 0;
        help_flag= -1;
        while ( argc > 0 ) {
            if ( strcmp( *argv, "+r" ) == 0 ) {
                options.rect_flag= 1;
                ++argv;
                --argc;
            }

            else if ( strcmp( *argv, "-r" ) == 0 ) {
                options.rect_flag= 0;
                ++argv;
                --argc;
            }

            else if ( strncmp( *argv, "+c", 2 ) == 0 ) {
                options.clip_flag= 1;
                options.clip_char= (*argv)[2];
                ++argv;
                --argc;
            }

            else if ( strcmp( *argv, "-c" ) == 0 ) {
                options.clip_flag= 0;
                ++argv;
                --argc;
            }

            else if ( strcmp( *argv, "+h" ) == 0 ) {
                help_flag= 1;
                ++argv;
                --argc;
            }

            else if ( strcmp( *argv, "-h" ) == 0 ) {
                help_flag= 0;
                ++argv;
                --argc;
            }

            else
                break;
        }

        /* handle help */
        if ( help_flag == 1 || ( argc == 0 && help_flag != 0 ) ) {
            printf( "%s version 1.2.1, (library %s%s found)\n\n", n, sn, snunit == 0? " not":"" );
            printf( "  This program converts LTSpice schematics into ASCII schematic output (or\n" );
            printf( "  files.)  If you specify no files at all, it accepts the LTSpice schematic\n" );
            printf( "  from its standard input device.  If you specify exactly one file, it dis-\n" );
            printf( "  plays the ASCII schematic output to the standard output device.  If you\n" );
            printf( "  specify more than one file, it then generates .TXT files otherwise having\n" );
            printf( "  the same name as the specified schematics.\n\n" );
            printf( "  These options are supported:\n    +h requests this help message, -h disables it.\n" );
            printf( "    +r enables rectangle drawing, -r disables it (default is -r)\n" );
            printf( "    +c enables clipboard copying, -c disables it (default is -c)\n" );
            printf( "    +c<char> enables clipboard copying and prepends <char> to each line\n\n" );
            printf( "  Usage:  %s <options> <filename> [<filename>]...\n", n );
            goto done;
        }

        /* winoldap verification, if clipboard requested */
        if ( options.clip_flag && !clpVerify( ) )
            options.clip_flag= 0;

        /* process the command line arguments as file names */
        switch ( argc ) {

        case 0:
            DisplayASCII( stdin, stdout, snunit );
            break;

        case 1:
            {
                FILE *f= fopen( argv[0], "r" );
                if ( f != NULL ) {
                    DisplayASCII( f, stdout, snunit );
                    fclose( f );
                }
            }
            break;

        default:
            options.clip_flag= 0;
            for ( i= 0; i < argc; ++i ) {
                FILE *fi= fopen( argv[i], "r" );
                if ( fi != NULL ) {
                    char *fn= BuildFullPath( argv[i], NULL, "TXT" );
                    FILE *fo= fopen( fn, "w" );
                    if ( fo != NULL ) {
                        DisplayASCII( fi, fo, snunit );
                        fclose( fo );
                    }
                    if ( fn != NULL )
                        free( fn );
                    fclose( fi );
                } else
                   fprintf( stderr, "%s: couldn't find %s\n", n, argv[i] );
            }
            break;
        }

done:
        /* close up the opened symbol file */
        if ( snunit != NULL )
            fclose( snunit );

        /* free up allocated strings */
        if ( sn != NULL )
            free( sn );
        if ( n != NULL )
            free( n );

    return 0;
}
